from copy import copy
from django import template
from django.template import Context
from django.template.loader import get_template
from django.utils.safestring import mark_safe
from django.contrib.admin.views.main import SEARCH_VAR
from suit.apps import ALL_FIELDS
from suit.compat import parse_qs

register = template.Library()


@register.simple_tag(takes_context=True)
def result_row_attrs(context, cl, row_index):
    """
    Returns row attributes based on object instance
    """
    row_index -= 1
    attrs = {
        'class': 'row1' if row_index % 2 == 0 else 'row2'
    }
    suit_row_attributes = getattr(cl.model_admin, 'suit_row_attributes', None)
    if not suit_row_attributes:
        return dict_to_attrs(attrs)

    instance = cl.result_list[row_index]
    new_attrs = suit_row_attributes(**{'obj': instance, 'request': context.request})
    if not new_attrs:
        return dict_to_attrs(attrs)

    # Validate
    if not isinstance(new_attrs, dict):
        raise TypeError('"suit_row_attributes" must return dict. Got: %s: %s' %
                        (new_attrs.__class__.__name__, new_attrs))

    # Merge 'class' attribute
    if 'class' in new_attrs:
        attrs['class'] += ' ' + new_attrs.pop('class')

    attrs.update(new_attrs)
    return dict_to_attrs(attrs)


@register.filter
def headers_handler(result_headers, cl):
    """
    Adds field name to css class, so we can style specific columns
    """
    # Django class attrib name
    attrib_key = 'class_attrib'

    suit_column_attributes = getattr(cl.model_admin, 'suit_column_attributes', None)
    if not suit_column_attributes:
        return result_headers

    for i, header in enumerate(result_headers):
        field_name = cl.list_display[i]
        attrs = copy(suit_column_attributes(field_name))
        if not attrs:
            continue

        # Validate
        if not isinstance(attrs, dict):
            raise TypeError('"suit_column_attributes" method must return dict. '
                            'Got: %s: %s' % (
                                attrs.__class__.__name__, attrs))

        classes = []
        defined_class = attrs.get('class')
        if defined_class:
            classes.append(defined_class)

        if attrib_key in header:
            existing_class = header[attrib_key].split('"')[1]
            if existing_class:
                classes.append(existing_class)

        if classes:
            attrs['class'] = ' '.join(classes)
            header[attrib_key] = dict_to_attrs(attrs)

    return result_headers


@register.filter
def cells_handler(results, cl):
    """
    Changes result cell attributes based on object instance and field name
    """
    suit_cell_attributes = getattr(cl.model_admin, 'suit_cell_attributes', None)
    if not suit_cell_attributes:
        return results

    class_pattern = 'class="'
    td_pattern = '<td'
    th_pattern = '<th'
    for row, result in enumerate(results):
        instance = cl.result_list[row]
        for col, item in enumerate(result):
            field_name = cl.list_display[col]
            attrs = copy(suit_cell_attributes(instance, field_name))
            if not attrs:
                continue

            # Validate
            if not isinstance(attrs, dict):
                raise TypeError('"suit_cell_attributes" method must return dict. '
                                'Got: %s: %s' % (
                                    attrs.__class__.__name__, attrs))

            # Merge 'class' attribute
            if class_pattern in item.split('>')[0] and 'class' in attrs:
                css_class = attrs.pop('class')
                replacement = '%s%s ' % (class_pattern, css_class)
                result[col] = mark_safe(
                    item.replace(class_pattern, replacement))

            # Add rest of attributes if any left
            if attrs:
                cell_pattern = td_pattern if item.startswith(
                    td_pattern) else th_pattern

                result[col] = mark_safe(
                    result[col].replace(cell_pattern,
                                        td_pattern + dict_to_attrs(attrs)))

    return results


def dict_to_attrs(attrs):
    return mark_safe(' ' + ' '.join(['%s="%s"' % (k, v)
                                     for k, v in attrs.items()]))


@register.inclusion_tag('suit/search_form.html')
def suit_search_form(cl):
    """
    Displays a search form for searching the list.
    """
    return {
        'cl': cl,
        'show_result_count': cl.result_count != cl.full_result_count,
        'search_var': SEARCH_VAR
    }


@register.simple_tag
def suit_admin_list_filter(cl, spec):
    if spec.template == 'admin/filter.html':
        spec.template = 'admin/filter_horizontal.html'

    tpl = get_template(spec.template)
    choices = list(spec.choices(cl))
    field_key = get_filter_id(spec)
    matched_key = field_key
    for choice in choices:
        query_string = choice['query_string'][1:]
        query_parts = parse_qs(query_string)

        value = ''
        matches = {}
        for key in query_parts.keys():
            if key == field_key:
                value = query_parts[key][0]
                matched_key = key
            elif key.startswith(
                            field_key + '__') or '__' + field_key + '__' in key:
                value = query_parts[key][0]
                matched_key = key

            if value:
                matches[matched_key] = value

        # Iterate matches, use first as actual values, additional for hidden
        i = 0
        for key, value in matches.items():
            if i == 0:
                choice['name'] = key
                choice['val'] = value
            else:
                choice['additional'] = '%s=%s' % (key, value)
            i += 1

    return tpl.render({
        'field_name': field_key,
        'title': spec.title,
        'choices': choices,
        'spec': spec,
    })


def _is_horizontal(horizontal_fields, field):
    return horizontal_fields == ALL_FIELDS or field in horizontal_fields

@register.filter
def suit_list_filter_vertical(filters, cl):
    filter_horizontal = getattr(cl.model_admin, 'suit_list_filter_horizontal', [])
    return [f for f in filters if not _is_horizontal(filter_horizontal, get_filter_id(f))]


@register.filter
def suit_list_filter_horizontal(filters, cl):
    filter_horizontal = getattr(cl.model_admin, 'suit_list_filter_horizontal', [])
    return [f for f in filters if _is_horizontal(filter_horizontal, get_filter_id(f))]


@register.filter
def suit_list_filter_horizontal_params(params, cl):
    # This collects params of vertical filters so they can be rendered as part of horizontal form
    excludes = ['_to_field', '_popup']
    vertical_keys = set()
    for vf in suit_list_filter_vertical(cl.filter_specs, cl):
        vertical_keys |= set(vf.expected_parameters())
    return [p for p in params if p[0] in vertical_keys or p[0] in excludes]


def get_filter_id(spec):
    try:
        return getattr(spec, 'field_path')
    except AttributeError:
        try:
            return getattr(spec, 'parameter_name')
        except AttributeError:
            pass
    return spec.title



from django import get_version
from django.apps import AppConfig
from django.contrib.admin.options import ModelAdmin
from . import VERSION

ALL_FIELDS = '__all__'

# Form row sizing as Bootstrap CSS grid classes: (for label, for field column)
SUIT_FORM_SIZE_LABEL = 'col-xs-12 col-sm-3 col-md-2'
SUIT_FORM_SIZE_INLINE = (SUIT_FORM_SIZE_LABEL, 'col-xs-12 col-sm-9 col-md-10 form-inline')
SUIT_FORM_SIZE_SMALL = (SUIT_FORM_SIZE_LABEL, 'col-xs-12 col-sm-6 col-md-5 col-lg-4')
SUIT_FORM_SIZE_HALF = (SUIT_FORM_SIZE_LABEL, 'col-xs-12 col-sm-7 col-md-6 col-lg-5')
SUIT_FORM_SIZE_LARGE = (SUIT_FORM_SIZE_LABEL, 'col-xs-12 col-sm-8 col-md-7 col-lg-6')
SUIT_FORM_SIZE_X_LARGE = (SUIT_FORM_SIZE_LABEL, 'col-xs-12 col-sm-9 col-md-8 col-lg-7')
SUIT_FORM_SIZE_XX_LARGE = (SUIT_FORM_SIZE_LABEL, 'col-xs-12 col-sm-9 col-md-10 col-lg-8')
SUIT_FORM_SIZE_XXX_LARGE = (SUIT_FORM_SIZE_LABEL, 'col-xs-12 col-sm-9 col-md-10 col-lg-9')
SUIT_FORM_SIZE_FULL = (SUIT_FORM_SIZE_LABEL, 'col-xs-12 col-sm-9 col-md-10')


class DjangoSuitConfig(AppConfig):
    name = 'suit'
    verbose_name = 'Django Suit'
    django_version = get_version()
    version = VERSION

    # Menu and header layout - horizontal or vertical
    layout = 'horizontal'

    # Set default list per page
    list_per_page = 20

    # Show changelist top actions only if any row is selected
    toggle_changelist_top_actions = True

    # Define menu
    #: :type: list of suit.menu.ParentItem
    menu = []

    # Automatically add home link
    menu_show_home = True

    # Define callback / handler to change menu before it is getting rendered
    menu_handler = None

    # Enables two column layout for change forms with submit row on the right
    form_submit_on_right = True

    # Hide name/"original" column for all tabular inlines.
    # May be overridden in Inline class by suit_form_inlines_hide_original = False
    form_inlines_hide_original = False

    # For size
    form_size = {
        'default': SUIT_FORM_SIZE_X_LARGE,
        # 'fields': {}
        'widgets': {
            'RelatedFieldWidgetWrapper': SUIT_FORM_SIZE_XXX_LARGE
        }
        # 'fieldsets': {}
    }

    # form_size setting can be overridden in ModelAdmin using suit_form_size parameter
    #
    # Example:
    # ----------------------------------------------
    # suit_form_size = {
    #     'default': 'col-xs-12 col-sm-2', 'col-xs-12 col-sm-10',
    #     'fields': {
    #          'field_name': SUIT_FORM_SIZE_LARGE,
    #          'field_name2': SUIT_FORM_SIZE_X_LARGE,
    #      },
    #      'widgets': {
    #          'widget_class_name': SUIT_FORM_SIZE_FULL,
    #          'AdminTextareaWidget': SUIT_FORM_SIZE_FULL,
    #      },
    #      'fieldsets': {
    #          'fieldset_name': SUIT_FORM_SIZE_FULL,
    #          'fieldset_name2': SUIT_FORM_SIZE_FULL,
    #      }
    # }

    def __init__(self, app_name, app_module):
        self.setup_model_admin_defaults()
        super(DjangoSuitConfig, self).__init__(app_name, app_module)

    def ready(self):
        super(DjangoSuitConfig, self).ready()

    def setup_model_admin_defaults(self):
        """
        Override some ModelAdmin defaults
        """
        if self.toggle_changelist_top_actions:
            ModelAdmin.actions_on_top = True
        ModelAdmin.actions_on_bottom = True

        if self.list_per_page:
            ModelAdmin.list_per_page = self.list_per_page



from django.contrib import admin
from django.db import models
from django.utils.safestring import mark_safe
try:
    from django.urls import reverse_lazy
except:
    from django.core.urlresolvers import reverse_lazy

"""
Adapted by using following examples:
https://djangosnippets.org/snippets/2887/
http://stackoverflow.com/a/7192721/641263
"""

link_to_prefix = 'link_to_'


def get_admin_url(instance, admin_prefix='admin', current_app=None):
    if not instance.pk:
        return
    return reverse_lazy(
        '%s:%s_%s_change' % (admin_prefix, instance._meta.app_label, instance._meta.model_name),
        args=(instance.pk,),
        current_app=current_app
    )


def get_related_field(name, short_description=None, admin_order_field=None, admin_prefix='admin'):
    """
    Create a function that can be attached to a ModelAdmin to use as a list_display field, e.g:
    client__name = get_related_field('client__name', short_description='Client')
    """
    as_link = name.startswith(link_to_prefix)
    if as_link:
        name = name[len(link_to_prefix):]
    related_names = name.split('__')

    def getter(self, obj):
        for related_name in related_names:
            if not obj:
                continue
            obj = getattr(obj, related_name)
        if obj and as_link:
            obj = mark_safe(u'<a href="%s" class="link-with-icon">%s<i class="fa fa-caret-right"></i></a>' % \
                            (get_admin_url(obj, admin_prefix, current_app=self.admin_site.name), obj))
        return obj

    getter.admin_order_field = admin_order_field or name
    getter.short_description = short_description or related_names[-1].title().replace('_', ' ')
    if as_link:
        getter.allow_tags = True
    return getter


class RelatedFieldAdminMetaclass(type(admin.ModelAdmin)):
    related_field_admin_prefix = 'admin'

    def __new__(cls, name, bases, attrs):
        new_class = super(RelatedFieldAdminMetaclass, cls).__new__(cls, name, bases, attrs)

        for field in new_class.list_display:
            if '__' in field or field.startswith(link_to_prefix):
                if not hasattr(new_class, field):
                    setattr(new_class, field, get_related_field(
                        field, admin_prefix=cls.related_field_admin_prefix))

        return new_class


class RelatedFieldAdmin(admin.ModelAdmin):
    """
    Version of ModelAdmin that can use linked and related fields in list_display, e.g.:
    list_display = ('link_to_user', 'address__city', 'link_to_address__city', 'address__country__country_code')
    """
    __metaclass__ = RelatedFieldAdminMetaclass

    def get_queryset(self, request):
        qs = super(RelatedFieldAdmin, self).get_queryset(request)

        # Include all related fields in queryset
        select_related = []
        for field in self.list_display:
            if '__' in field:
                if field.startswith(link_to_prefix):
                    field = field[len(link_to_prefix):]
                select_related.append(field.rsplit('__', 1)[0])

        # Include all foreign key fields in queryset.
        # This is based on ChangeList.get_query_set().
        # We have to duplicate it here because select_related() only works once.
        # Can't just use list_select_related because we might have multiple__depth__fields it won't follow.
        model = qs.model
        for field_name in self.list_display:
            try:
                field = model._meta.get_field(field_name)
            except models.FieldDoesNotExist:
                continue

            if isinstance(field.remote_field, models.ManyToOneRel):
                select_related.append(field_name)

        return qs.select_related(*select_related)



from os.path import dirname, join, abspath
from django.template.loaders.filesystem import Loader as FilesystemLoader

_cache = {}


class Loader(FilesystemLoader):
    is_usable = True

    def get_template_sources(self, template_name, template_dirs=None):
        """
        Returns the absolute paths to "template_name" in the specified app.
        If the name does not contain an app name (no colon), an empty list
        is returned.
        The parent FilesystemLoader.load_template_source() will take care
        of the actual loading for us.
        """
        if not ':' in template_name:
            return []
        app_name, template_name = template_name.split(":", 1)
        template_dir = get_app_template_dir(app_name)
        if template_dir:
            return [join(template_dir, template_name)]
        else:
            return []


def get_app_template_dir(app_name):
    """Get the template directory for an application

    We do not use django.db.models.get_app, because this will fail if an
    app does not have any models.

    Returns a full path, or None if the app was not found.
    """
    from django.conf import settings
    from importlib import import_module
    if app_name in _cache:
        return _cache[app_name]
    template_dir = None
    for app in settings.INSTALLED_APPS:
        if app.split('.')[-1] == app_name:
            # Do not hide import errors; these should never happen at this point
            # anyway
            mod = import_module(app)
            template_dir = join(abspath(dirname(mod.__file__)), 'templates')
            break
    _cache[app_name] = template_dir
    return template_dir



from copy import deepcopy
from django.utils.translation import ugettext_lazy as _


class ChildItem(object):
    def __init__(self, label=None, model=None, url=None, target_blank=False, permissions=None):
        self.label = label
        self.model = model
        self.url = url
        self.target_blank = target_blank
        self.permissions = permissions
        self.is_active = None
        self._is_forbidden = False
        self._parent_item = None
        self._url_name = None

    def _key(self):
        if self._parent_item and self.model and not '.' in self.model:
            return '.'.join([self._parent_item._key(), self.model])
        return self.model


class ParentItem(ChildItem):
    def __init__(self, label=None, app=None, url=None, target_blank=False, permissions=None,
                 children=None, align_right=False, use_first_child_url=True, icon=None):
        super(ParentItem, self).__init__(label, None, url, target_blank, permissions)
        self.user_children = children or []
        self.children = []
        self.align_right = align_right
        self.icon = icon
        self.app = app
        self.use_first_child_url = use_first_child_url

    def _key(self):
        return self.app


class MenuManager(object):
    def __init__(self, available_apps, context, request):
        from .config import get_config_instance, get_current_app

        super(MenuManager, self).__init__()

        # Variable available_apps structure:
        # https://docs.djangoproject.com/en/1.9/ref/contrib/admin/#adminsite-methods
        self.available_apps = available_apps

        self.context = context
        self.request = request
        self.current_app = get_current_app(request)
        self.suit_config = get_config_instance(self.current_app)
        self.user_menu = self.suit_config.menu
        self.menu_items = None
        self.aligned_right_menu_items = []
        self.active_parent_item = None
        self._available_apps = {'apps': {}, 'models': {}}

    def __iter__(self):
        for each in self.get_menu_items():
            yield each

    def get_menu_items(self):
        if self.menu_items is None:
            self.menu_items = self.build_menu()
            if self.suit_config.menu_handler:
                if not callable(self.suit_config.menu_handler):
                    raise TypeError('Django Suit "menu_handler" must callable')
                self.menu_items = self.suit_config.menu_handler(
                    self.menu_items, self.request, self.context)

        return self.menu_items

    def build_menu(self):
        if not self.user_menu:
            self.map_native_apps()
            return self.mark_active(self.build_menu_by_available_apps())

        self.map_native_apps()

        menu_items = []
        user_menu = deepcopy(self.user_menu)
        for parent_item in user_menu:
            native_app = self.find_native_app(parent_item)
            if parent_item.user_children:
                for child_item in parent_item.user_children:
                    child_item._parent_item = parent_item
                    native_model = self.find_native_model(native_app, child_item)
                    self.handle_child_menu(child_item, native_model)

                    # Add allowed children
                    if not child_item._is_forbidden:
                        parent_item.children.append(child_item)

            elif native_app:
                self.make_children_from_native_app(parent_item, native_app)

            self.handle_parent_menu(parent_item, native_app)

            # Add allowed items
            if not self.parent_item_is_forbidden(parent_item, native_app):
                menu_items.append(parent_item)

                if parent_item.align_right and self.suit_config.layout == 'horizontal':
                    self.aligned_right_menu_items.append(parent_item)

        if self.suit_config.menu_show_home:
            home_item = ParentItem(_('Home'), url='admin:index', icon='fa fa-home')
            menu_items.insert(0, self.handle_user_url(home_item))

        return self.mark_active(menu_items)

    def map_native_apps(self):
        """
        Make dictionary of native apps and models for easier matching
        """
        for native_app in self.available_apps:
            app_key = native_app['app_url'].split('/')[-2]
            self._available_apps['apps'][app_key] = native_app
            for native_model in native_app['models']:
                if 'admin_url' not in native_model:
                    # Can happen with incomplete permissions, like Delete only, etc.
                    continue
                model_key = '.'.join(native_model['admin_url'].split('/')[-3:-1])
                native_model['model'] = model_key
                self._available_apps['models'][model_key] = native_model
                model_key2 = '.'.join([app_key, native_model['object_name'].lower()])
                self._available_apps['models'][model_key2] = native_model

    def find_native_app(self, parent_item):
        """
        :type parent_item: ParentItem
        """
        return self._available_apps['apps'].get(parent_item._key())

    def find_native_model(self, native_app, child_item):
        """
        :type native_app: dict
        :type child_item: ChildItem
        """
        return self._available_apps['models'].get(child_item._key())

    def build_menu_by_available_apps(self):
        menu_items = []
        for native_app in self.available_apps:
            parent_item = self.make_parent_from_native_app(native_app)
            menu_items.append(parent_item)
            self.make_children_from_native_app(parent_item, native_app)
            if parent_item.children:
                parent_item.url = parent_item.children[0].url
        return menu_items

    def make_parent_from_native_app(self, native_app):
        """
        :type native_app: dict
        """
        parent_item = ParentItem(native_app['name'], url=native_app['app_url'])
        return parent_item

    def make_children_from_native_app(self, parent_item, native_app):
        """
        :type parent_item: ParentItem
        :type native_app: dict
        """
        for native_model in native_app['models']:
            child_item = self.make_child_from_native_model(native_model)
            parent_item.children.append(child_item)
            child_item._parent_item = parent_item

    def make_child_from_native_model(self, native_model):
        """
        :type native_model: dict
        """
        child_item = ChildItem(native_model['name'],  model=native_model.get('model'), url=native_model['admin_url'])
        return child_item

    def handle_parent_menu(self, parent_item, native_app):
        """
        :type parent_item: ParentItem
        :type native_app: dict
        """
        if not parent_item.label:
            parent_item.label = native_app['name'] if native_app else 'Untitled'
        if not parent_item.url:
            if parent_item.children and parent_item.use_first_child_url:
                parent_item.url = parent_item.children[0].url
            elif native_app:
                parent_item.url = native_app['app_url']
        else:
            self.handle_user_url(parent_item)

    def handle_child_menu(self, child_item, native_model):
        """
        :type child_item: ChildItem
        :type native_model: dict
        """
        # Handle permissions
        if self.child_item_is_forbidden(child_item, native_model):
            return

        # Handle label
        if not child_item.label:
            child_item.label = native_model['name'] if native_model else 'Untitled'

        # Handle URL
        if not child_item.url:
            if native_model:
                child_item.url = native_model['admin_url']
        else:
            self.handle_user_url(child_item)
        if not child_item.url:
            child_item.url = '#not-found'

    def handle_user_url(self, menu_item):
        """
        Evaluate user defined URL
        :type menu_item: ChildItem or ParentItem
        """
        if callable(menu_item.url):
            menu_item.url = menu_item.url(self.request, self.context)
            return menu_item
        if '/' in menu_item.url:
            return menu_item
        try:
            from django.urls import reverse, NoReverseMatch
        except:
            from django.core.urlresolvers import reverse, NoReverseMatch
        try:
            menu_item.url = reverse(menu_item.url, current_app=self.current_app)
            menu_item._url_name = menu_item.url
        except NoReverseMatch:
            pass
        return menu_item

    def parent_item_is_forbidden(self, parent_item, native_app):
        """
        :type parent_item: ParentItem
        """
        # Set as forbidden if model param specified but native_model not found
        if not parent_item.url and not parent_item.children:
            parent_item._is_forbidden = True
            return True

        return self.item_is_forbidden_by_custom_permissions(parent_item)

    def child_item_is_forbidden(self, child_item, native_model):
        """
        :type child_item: ChildItem
        """
        # Set as forbidden if model param specified but native_model not found
        if not child_item.url and not native_model:
            child_item._is_forbidden = True
            return True

        return self.item_is_forbidden_by_custom_permissions(child_item)

    def item_is_forbidden_by_custom_permissions(self, menu_item):
        """
        :type menu_item: ChildItem or ParentItem
        """
        if menu_item.permissions and \
                not self.user_has_permission(menu_item.permissions):
            menu_item._is_forbidden = True
            return True

    def user_has_permission(self, perms):
        perms = perms if isinstance(perms, (list, tuple)) else (perms,)
        return self.request.user.has_perms(perms)

    def mark_active(self, menu_items):
        active_child, active_child_by_url = None, None
        active_parent, active_parent_by_url = None, None

        # Represents: %s.%s % (opts.app_label, opts.app_model_name)
        opts_key = str(self.context.get('opts'))
        url_name = self.context.get('url_name')

        request_path = str(self.request.path)

        for parent_item in menu_items:
            if not active_child:
                for child_item in parent_item.children:
                    if opts_key == child_item._key():
                        active_child = child_item
                        break
                    elif not active_child_by_url and request_path == child_item.url:
                        active_child_by_url = child_item

            if active_child:
                break

            if not active_parent:
                if url_name and url_name == parent_item._url_name or request_path == parent_item.url:
                    active_parent = parent_item

        if not active_child:
            active_child = active_child_by_url

        if active_child:
            active_child.is_active = True
            active_parent = active_child._parent_item

        if active_parent:
            active_parent.is_active = True
            self.active_parent_item = active_parent

        return menu_items
